from ..compat_typing import Literal
from ..http_request import HttpRequestParameters, request


def get_object_source(
    http_request_parameters: HttpRequestParameters,
    object_uri: str,
    version: Literal["active", "inactive"] = "active",
) -> str:

    params = {}
    if version:
        params["version"] = version

    response = request(
        http_request_parameters=http_request_parameters,
        uri=object_uri,
        method="GET",
        body="",
        params=params,
        content_type="application/xml",
    )

    if response.status_code == 200:
        return response.text
    else:
        raise Exception(
            f"{response.status_code} - Failed to get object source.\n{response.text}"
        )


def set_object_source(
    http_request_parameters: HttpRequestParameters,
    object_uri: str,
    source_code: str,
    lock_handle: str,
) -> bool:
    response = request(
        http_request_parameters=http_request_parameters,
        uri=object_uri,
        method="PUT",
        body=source_code,
        params={"lockHandle": lock_handle},
        content_type="text/plain; charset=utf-8",
    )

    if response.status_code == 200:
        return True
    else:
        raise Exception(
            f"{response.status_code} - Failed to set source code for {object_uri}\n{response.text}"
        )



from ..http_request import HttpRequestParameters, request


def delete(http_request_parameters: HttpRequestParameters, object_uri: str, lock_handle: str) -> bool:
    response = request(
        http_request_parameters=http_request_parameters,
        uri=object_uri,
        body="",
        params={"lockHandle": lock_handle},
        method="DELETE",
    )
    if response.status_code == 200:
        return True
    else:
        raise Exception(
            f"{response.status_code} Failed to delete {object_uri}.\n{response.text}"
        )


import datetime

import pytest
from abap_adt_py.adt_client import AdtClient

TEST_CREATE_ORDER = 1
TEST_CREATE_CLASS_INTERFACE_ORDER = 2


@pytest.mark.order(TEST_CREATE_ORDER)
def test_create_report(client: AdtClient, test_report: str):
    file_created = client.create(
        object_type="PROG/P",
        name=test_report,
        description="Test Program",
        parent="$TMP",
    )
    assert file_created


@pytest.mark.order(TEST_CREATE_ORDER)
def test_create_class(client: AdtClient, test_class: str):
    file_created = client.create(
        object_type="CLAS/OC",
        name=test_class,
        description="Test Class",
        parent="$TMP",
    )
    assert file_created


@pytest.mark.order(TEST_CREATE_CLASS_INTERFACE_ORDER)
def test_create_abap_testclass_include(client: AdtClient, test_class: str):
    uri = f"/sap/bc/adt/oo/classes/{test_class}"
    lock_handle = client.lock(uri)
    file_created = client.create_test_class_include(test_class, lock_handle)
    client.unlock(uri, lock_handle)
    assert file_created


# @pytest.mark.order(TEST_CREATE_ORDER)
# def test_create_table(client: AdtClient, test_table: str):
#     file_created = client.create(
#         object_type="TABL/DT",
#         name=test_table,
#         description="Test Table",
#         parent="$TMP",
#     )
#     assert file_created


# @pytest.mark.order(TEST_CREATE_ORDER)
# def test_create_interface(client: AdtClient, test_interface: str):
#     file_created = client.create(
#         object_type="INTF/OI",
#         name=test_interface,
#         description="Test Interface",
#         parent="$TMP",
#     )
#     assert file_created


# @pytest.mark.order(TEST_CREATE_ORDER)
# def test_create_executable_program(client: AdtClient, test_executable: str):
#     file_created = client.create(
#         object_type="PROG/I",
#         name=test_executable,
#         description="Test Executable Program",
#         parent="$TMP",
#     )
#     assert file_created


# @pytest.mark.order(TEST_CREATE_ORDER)
# def test_create_function_group(client: AdtClient, test_fugr: str):
#     file_created = client.create(
#         object_type="FUGR/F",
#         name=test_fugr,
#         description="Test Function Group",
#         parent="$TMP",
#     )
#     assert file_created


# @pytest.mark.order(TEST_CREATE_ORDER)
# def test_create_function_module(client: AdtClient, test_func_mod: str):
#     file_created = client.create(
#         object_type="FUGR/FF",
#         name=test_func_mod,
#         description="Test Function Module",
#         parent="$TMP",
#     )
#     assert file_created


# @pytest.mark.order(TEST_CREATE_ORDER)
# def test_create_message_class(client: AdtClient, test_msag: str):
#     file_created = client.create(
#         object_type="MSAG/N",
#         name=test_msag,
#         description="Test Message Class",
#         parent="$TMP",
#     )
#     assert file_created


# @pytest.mark.order(TEST_CREATE_ORDER)
# def test_create_dcl_source(client: AdtClient, test_dcls: str):
#     file_created = client.create(
#         object_type="DCLS/DL",
#         name=test_dcls,
#         description="Test DCL Source",
#         parent="$TMP",
#     )
#     assert file_created


# @pytest.mark.order(TEST_CREATE_ORDER)
# def test_create_ddls_view(client: AdtClient, test_ddls: str):
#     file_created = client.create(
#         object_type="DDLS/DF",
#         name=test_ddls,
#         description="Test CDS View",
#         parent="$TMP",
#     )
#     assert file_created


# @pytest.mark.order(TEST_CREATE_ORDER)
# def test_create_ddlx_extension(client: AdtClient, test_ddlx: str):
#     file_created = client.create(
#         object_type="DDLX/EX",
#         name=test_ddlx,
#         description="Test CDS View Extension",
#         parent="$TMP",
#     )
#     assert file_created


# @pytest.mark.order(TEST_CREATE_ORDER)
# def test_create_data_element(client: AdtClient, test_dtel: str):
#     file_created = client.create(
#         object_type="DTEL/DE",
#         name=test_dtel,
#         description="Test Data Element",
#         parent="$TMP",
#     )
#     assert file_created



from ..http_request import HttpRequestParameters, request
from ..response_parsing import find_xml_element_text


def lock(http_request_parameters: HttpRequestParameters, object_uri: str) -> str:
    response = request(
        http_request_parameters=http_request_parameters,
        uri=object_uri,
        body="",
        params={"_action": "LOCK", "accessMode": "MODIFY"},
        method="POST",
    )
    if response.status_code == 200:
        lock_handle = find_xml_element_text(response.text, ".//LOCK_HANDLE")
        return lock_handle
    else:
        raise Exception(
            f"{response.status_code} Failed to lock {object_uri}.\n{response.text}"
        )


def unlock(
    http_request_parameters: HttpRequestParameters, object_uri: str, lock_handle: str
) -> bool:
    response = request(
        http_request_parameters=http_request_parameters,
        uri=object_uri,
        body="",
        params={"_action": "UNLOCK", "lockHandle": lock_handle},
        method="POST",
        content_type="plain/text; charset=utf-8",
    )
    if response.status_code == 200:
        return True
    else:
        raise Exception(
            f"{response.status_code} Failed to unlock {object_uri}\n{response.text}"
        )



from ..compat_typing import Literal, NotRequired, TypedDict
from ..http_request import HttpRequestParameters, request


class PrettyPrintSettings(TypedDict):
    indentation: bool
    style: Literal[
        "toLower", "toUpper", "keywordLower", "keywordUpper", "none", "keywordAuto"
    ]
    condense_method_calls: NotRequired[bool]
    default_auto_style: NotRequired[Literal["keywordLower", "keywordUpper"]]
    format_method_calls: NotRequired[bool]
    keep_identifier: NotRequired[bool]


def set_pretty_printer_settings(
    http_request_parameters: HttpRequestParameters, settings: PrettyPrintSettings
):

    condense_method_calls = (
        f"prettyprintersettings:condenseMethodCalls=\"{str(settings['condense_method_calls']).lower()}\""
        if "condense_method_calls" in settings
        else ""
    )
    default_auto_style = (
        f"prettyprintersettings:defaultAutoStyle=\"{settings['default_auto_style']}\""
        if "default_auto_style" in settings
        else ""
    )
    format_method_calls = (
        f"prettyprintersettings:formatMethodCalls=\"{str(settings['format_method_calls']).lower()}\""
        if "format_method_calls" in settings
        else ""
    )
    indentation = (
        f"prettyprintersettings:indentation=\"{str(settings['indentation']).lower()}\""
        if "indentation" in settings
        else ""
    )
    keep_identifier = (
        f"prettyprintersettings:keepIdentifier=\"{str(settings['keep_identifier']).lower()}\""
        if "keep_identifier" in settings
        else ""
    )
    style = (
        f"prettyprintersettings:style=\"{settings['style']}\""
        if "style" in settings
        else ""
    )

    body = f"""<?xml version="1.0" encoding="UTF-8"?>
    <prettyprintersettings:PrettyPrinterSettings 
    xmlns:prettyprintersettings="http://www.sap.com/adt/prettyprintersettings"
        {condense_method_calls}
        {default_auto_style}
        {format_method_calls}
        {indentation}
        {keep_identifier}
        {style}
        />
        """

    response = request(
        http_request_parameters,
        uri="/sap/bc/adt/abapsource/prettyprinter/settings",
        params={},
        body=body,
        method="PUT",
        content_type="application/xml",
    )
    if response.status_code == 200:
        return True
    else:
        raise Exception(
            f"{response.status_code} Failed to set prettyprint settings\n{response.text}"
        )


def prettyprint(http_request_parameters: HttpRequestParameters, src: str):
    response = request(
        http_request_parameters,
        uri="/sap/bc/adt/abapsource/prettyprinter",
        params={},
        body=src,
        method="POST",
        content_type="text/plain; charset=utf-8",
    )
    if response.status_code == 200:
        return response.text
    else:
        raise Exception(
            f"{response.status_code} Failed to prettyprint\n{response.text}"
        )



from ..http_request import HttpRequestParameters, request
from ..response_parsing import find_xml_elements_attributes
from ..compat_typing import List, Dict


def search_object(
    http_request_parameters: HttpRequestParameters, query: str, max_results: int = 1
) -> List[Dict[str, str]]:

    response = request(
        http_request_parameters=http_request_parameters,
        uri="/sap/bc/adt/repository/informationsystem/search",
        method="GET",
        body="",
        params={"operation": "quickSearch", "query": query, "maxResults": max_results},
    )
    elements = find_xml_elements_attributes(response.text, "adtcore:objectReference")
    return elements



from ..http_request import HttpRequestParameters, request
from ..response_parsing import find_xml_element_attributes, find_xml_elements_attributes


def activate(
    http_request_parameters: HttpRequestParameters, object_name: str, object_uri: str
) -> bool:

    body = f"""
    <?xml version="1.0" encoding="UTF-8"?>
    <adtcore:objectReferences xmlns:adtcore="http://www.sap.com/adt/core">
        <adtcore:objectReference adtcore:uri="{object_uri}" adtcore:name="{object_name}"/>
    </adtcore:objectReferences>
    """

    response = request(
        http_request_parameters,
        uri="/sap/bc/adt/activation",
        params={"method": "activate", "preauditRequested": "true"},
        body=body,
        method="POST",
        content_type="application/xml",
    )
    properties = find_xml_element_attributes(response.text, "chkl:properties")
    if (
        properties["activationExecuted"] == "true"
        or properties["generationExecuted"] == "true"
    ):
        return True
    else:
        msg_elements = find_xml_elements_attributes(response.text, "msg")
        raise Exception(f"{response.status_code} - Activation failed.\n{msg_elements}")



import base64
import xml.etree.ElementTree as et
from ..compat_typing import TypedDict, Literal, NotRequired
from ..http_request import HttpRequestParameters, request
from .xml_namespaces import XML_NAMESPACES


class SyntaxCheckResult(TypedDict):
    uri: str
    line: NotRequired[int]
    offset: NotRequired[int]
    type: str
    short_text: str


def _parse_syntax_check_response(response_text: str) -> list[SyntaxCheckResult]:
    root = et.fromstring(response_text)
    messages = []
    for msg in root.findall(".//chkrun:checkMessage", XML_NAMESPACES):
        uri: str = msg.get(f'{{{XML_NAMESPACES["chkrun"]}}}uri', "")

        if "#start=" in uri:
            line, offset = uri[uri.index("#start=") + 7 :].split(",")
            line, offset = int(line), int(offset)
            uri = uri[: uri.index("#start=")]
        else:
            line, offset = None, None

        uri = uri.removesuffix("/source/main")

        type = msg.get(f'{{{XML_NAMESPACES["chkrun"]}}}type', "")
        short_text = msg.get(f'{{{XML_NAMESPACES["chkrun"]}}}shortText', "")

        message: SyntaxCheckResult = {
            "uri": uri,
            "type": type,
            "short_text": short_text,
        }

        if line and offset:
            message["line"] = line
            message["offset"] = offset

        messages.append(message)
    return messages


def syntax_check(
    http_request_parameters: HttpRequestParameters,
    object_uri: str,
    include_uri: str,
    source_code: str,
    version: Literal["active", "inactive"] = "active",
) -> list[SyntaxCheckResult]:
    body = f"""
    <?xml version="1.0" encoding="UTF-8"?>
    <chkrun:checkObjectList xmlns:chkrun="http://www.sap.com/adt/checkrun" xmlns:adtcore="http://www.sap.com/adt/core">
    <chkrun:checkObject adtcore:uri="{object_uri}" chkrun:version="{version}">
        <chkrun:artifacts>
        <chkrun:artifact chkrun:contentType="text/plain; charset=utf-8" chkrun:uri="{include_uri}">
            <chkrun:content>{base64.b64encode(source_code.encode("utf-8")).decode("utf-8")}</chkrun:content>
        </chkrun:artifact>
        </chkrun:artifacts>
    </chkrun:checkObject>
    </chkrun:checkObjectList>
    """

    response = request(
        http_request_parameters,
        uri="/sap/bc/adt/checkruns?reporters=abapCheckRun",
        params={},
        body=body,
        method="POST",
        content_type="application/*",
    )
    if 200 <= response.status_code < 300:
        messages = _parse_syntax_check_response(response.text)
        return messages
    else:
        raise Exception(
            f"{response.status_code} - Failed to check syntax for {object_uri}\n{response.text}"
        )



import requests
from requests.auth import HTTPBasicAuth

from .compat_typing import Literal, List, Dict
from .api.syntax import SyntaxCheckResult, syntax_check
from .api.objectstructure import object_structure
from .api.prettyprint import (
    PrettyPrintSettings,
    prettyprint,
    set_pretty_printer_settings,
)
from .api.create import create, create_test_class_include
from .api.activate import activate
from .api.create import ObjectTypes
from .api.delete import delete
from .api.lock import lock, unlock
from .api.login import login
from .api.content import get_object_source, set_object_source
from .api.search import search_object
from .api.unittest import UnitTestAlert, UnittestFlags, run_unit_test
from .http_request import HttpRequestParameters


class AdtClient:
    sap_host: str
    csrf_token: str = "fetch"
    request_number: int = 0
    statefulness: Literal["stateless", "stateful"] = "stateless"

    def __init__(
        self, sap_host: str, username: str, password: str, client: str, language: str
    ):
        self.username = username
        self.session = requests.Session()
        self.session.auth = HTTPBasicAuth(username, password)
        self.sap_host = sap_host
        self.client = client
        self.language = language

    def build_request_parameters(self) -> HttpRequestParameters:
        http_request_parameters: HttpRequestParameters = {
            "host": self.sap_host,
            "csrf_token": self.csrf_token,
            "statefulness": self.statefulness,
            "request_number": self.request_number,
            "session": self.session,
        }
        self.request_number += 1
        return http_request_parameters

    def login(self) -> bool:
        http_request_parameters = self.build_request_parameters()
        csrf_token = login(http_request_parameters)
        if csrf_token:
            self.csrf_token = csrf_token
            return True
        else:
            raise Exception("Login failed.")

    def search_object(self, query: str, max_results: int = 1) -> List[Dict[str, str]]:
        http_request_parameters = self.build_request_parameters()
        elements = search_object(http_request_parameters, query, max_results)
        return elements

    def get_object_source(
        self, object_uri: str, version: Literal["active", "inactive"] = "active"
    ) -> str:
        http_request_parameters = self.build_request_parameters()
        response = get_object_source(http_request_parameters, object_uri, version)
        return response

    def activate(self, object_name: str, object_uri: str) -> bool:
        http_request_parameters = self.build_request_parameters()
        response = activate(http_request_parameters, object_name, object_uri)
        return response

    def lock(self, object_uri: str) -> str:
        self.statefulness = "stateful"
        http_request_parameters = self.build_request_parameters()
        response = lock(http_request_parameters, object_uri)
        return response

    def unlock(self, object_uri: str, lock_handle: str) -> bool:
        http_request_parameters = self.build_request_parameters()
        response = unlock(http_request_parameters, object_uri, lock_handle)
        self.statefulness = "stateless"
        return response

    def set_object_source(
        self, object_uri: str, source_code: str, lock_handle: str
    ) -> bool:
        http_request_parameters = self.build_request_parameters()
        response = set_object_source(
            http_request_parameters, object_uri, source_code, lock_handle
        )
        return response

    def run_unit_test(
        self, object_uri: str, unit_test_flags: UnittestFlags = UnittestFlags()
    ) -> List[UnitTestAlert]:
        http_request_parameters = self.build_request_parameters()
        response = run_unit_test(http_request_parameters, object_uri, unit_test_flags)
        return response

    def delete(self, object_uri: str, lock_handle: str) -> bool:
        http_request_parameters = self.build_request_parameters()
        response = delete(http_request_parameters, object_uri, lock_handle)
        return response

    def create(
        self, object_type: ObjectTypes, name: str, parent: str, description: str
    ) -> bool:
        http_request_parameters = self.build_request_parameters()
        response = create(
            http_request_parameters,
            object_type,
            name,
            parent,
            description,
            self.username,
        )
        return response

    def create_test_class_include(self, class_name: str, lock_handle: str) -> bool:
        http_request_parameters = self.build_request_parameters()
        response = create_test_class_include(
            http_request_parameters, class_name, lock_handle
        )
        return response

    def prettyprint(self, src: str) -> str:
        http_request_parameters = self.build_request_parameters()
        response = prettyprint(http_request_parameters, src)
        return response

    def prettyprint_settings(self, settings: PrettyPrintSettings) -> bool:
        http_request_parameters = self.build_request_parameters()
        response = set_pretty_printer_settings(http_request_parameters, settings)
        return response

    def syntax_check(
        self,
        object_uri: str,
        include_uri: str,
        src: str,
        version: Literal["active", "inactive"] = "active",
    ) -> List[SyntaxCheckResult]:
        http_request_parameters = self.build_request_parameters()
        response = syntax_check(
            http_request_parameters, object_uri, include_uri, src, version
        )
        return response

    def object_structure(self, object_uri: str):
        http_request_parameters = self.build_request_parameters()
        response = object_structure(http_request_parameters, object_uri)
        return response



from typing import Optional, Dict, List  # Always available since Python 3.5

try:
    # Python 3.11+ where all are in typing
    from typing import Literal, NotRequired, TypeAlias, TypedDict
except ImportError:
    try:
        # Python 3.8 - 3.10: Literal in typing, others from typing_extensions
        from typing import Literal
        from typing_extensions import NotRequired, TypeAlias, TypedDict
    except ImportError:
        # Python 3.7 fallback: all from typing_extensions
        from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict



from ..compat_typing import Literal, TypeAlias, Dict, TypedDict
from ..http_request import HttpRequestParameters, request


class CreateableTypeDetails(TypedDict):
    path: str
    xml_name: str
    xml_namespace: str


ObjectTypes: TypeAlias = Literal[
    "PROG/P",
    "CLAS/OC",
    "TABL/DT",
    "INTF/OI",
    "PROG/I",
    "FUGR/F",
    "FUGR/FF",
    "MSAG/N",
    "DCLS/DL",
    "DDLS/DF",
    "DDLX/EX",
    "DTEL/DE",
]

CreateableTypes: TypeAlias = Dict[ObjectTypes, CreateableTypeDetails]
CREATEABLE_TYPES: CreateableTypes = {
    "PROG/P": {
        "path": "/sap/bc/adt/programs/programs",
        "xml_name": "program:abapProgram",
        "xml_namespace": 'xmlns:program="http://www.sap.com/adt/programs/programs"',
    },
    "CLAS/OC": {
        "path": "/sap/bc/adt/oo/classes",
        "xml_name": "class:abapClass",
        "xml_namespace": 'xmlns:class="http://www.sap.com/adt/oo/classes"',
    },
    "TABL/DT": {
        "path": "/sap/bc/adt/ddic/tables",
        "xml_name": "blue:blueSource",
        "xml_namespace": 'xmlns:blue="http://www.sap.com/wbobj/blue"',
    },
    "INTF/OI": {
        "path": "/sap/bc/adt/oo/interfaces",
        "xml_name": "intf:abapInterface",
        "xml_namespace": 'xmlns:intf="http://www.sap.com/adt/oo/interfaces"',
    },
    "PROG/I": {
        "path": "/sap/bc/adt/programs/includes",
        "xml_name": "include:abapInclude",
        "xml_namespace": 'xmlns:include="http://www.sap.com/adt/programs/includes"',
    },
    "FUGR/F": {
        "path": "/sap/bc/adt/functions/groups",
        "xml_name": "group:abapFunctionGroup",
        "xml_namespace": 'xmlns:group="http://www.sap.com/adt/functions/groups"',
    },
    "FUGR/FF": {
        "path": "/sap/bc/adt/functions/groups/{}/fmodules",
        "xml_name": "fmodule:abapFunctionModule",
        "xml_namespace": 'xmlns:fmodule="http://www.sap.com/adt/functions/fmodules"',
    },
    "MSAG/N": {
        "path": "/sap/bc/adt/messageclass",
        "xml_name": "mc:messageClass",
        "xml_namespace": 'xmlns:mc="http://www.sap.com/adt/MessageClass"',
    },
    "DCLS/DL": {
        "path": "/sap/bc/adt/acm/dcl/sources",
        "xml_name": "dcl:dclSource",
        "xml_namespace": 'xmlns:dcl="http://www.sap.com/adt/acm/dclsources"',
    },
    "DDLS/DF": {
        "path": "/sap/bc/adt/ddic/ddl/sources",
        "xml_name": "ddl:ddlSource",
        "xml_namespace": 'xmlns:ddl="http://www.sap.com/adt/ddic/ddlsources"',
    },
    "DDLX/EX": {
        "path": "/sap/bc/adt/ddic/ddlx/sources",
        "xml_name": "ddlx:ddlxSource",
        "xml_namespace": 'xmlns:ddlx="http://www.sap.com/adt/ddic/ddlxsources"',
    },
    "DTEL/DE": {
        "path": "/sap/bc/adt/ddic/dataelements",
        "xml_name": "blue:wbobj",
        "xml_namespace": 'xmlns:blue="http://www.sap.com/wbobj/dictionary/dtel"',
    },
}


def _build_body(
    description: str, name: str, package: str, owner: str, object_type: ObjectTypes
) -> str:

    if object_type in ["FUGR/FF"]:
        parent_ref = f"""
        <adtcore:containerRef adtcore:name="{owner}" 
            adtcore:type="FUGR/F"
            adtcore:uri="{package}" />"""
    else:
        parent_ref = f"""
        <adtcore:packageRef adtcore:name="{package}"/>
        """

    creatable_object = CREATEABLE_TYPES[object_type]

    return f"""
    <?xml version="1.0" encoding="UTF-8"?>
    <{creatable_object["xml_name"]} {creatable_object["xml_namespace"]}
        xmlns:adtcore="http://www.sap.com/adt/core"
        adtcore:description="{description}"
        adtcore:name="{name}" adtcore:type="{object_type}"
        adtcore:responsible="{owner}" >
    {parent_ref}
    </{creatable_object["xml_name"]}>
    """


def create(
    http_request_parameters: HttpRequestParameters,
    object_type: ObjectTypes,
    name: str,
    parent: str,
    description: str,
    owner: str,
) -> bool:

    object_uri = CREATEABLE_TYPES[object_type]["path"]

    # noop for most objects but eg. function modules need their function group in their object uri
    object_uri = object_uri.format(parent)

    body = _build_body(
        description=description,
        name=name,
        package=parent,
        object_type=object_type,
        owner=owner,
    )

    response = request(
        http_request_parameters=http_request_parameters,
        uri=object_uri,
        method="POST",
        body=body,
        params={},
        content_type="application/*",
    )

    if 200 <= response.status_code <= 300:
        return True
    else:
        raise Exception(
            f"{response.status_code} - Failed to create object {name}\n{response.text}"
        )


def create_test_class_include(
    http_request_parameters: HttpRequestParameters, class_name: str, lock_handle: str
) -> bool:

    body = f"""
        <?xml version="1.0" encoding="UTF-8"?>
        <class:abapClassInclude xmlns:class="http://www.sap.com/adt/oo/classes"
            xmlns:adtcore="http://www.sap.com/adt/core" adtcore:name="{class_name}" class:includeType="testclasses"/>
        """

    response = request(
        http_request_parameters=http_request_parameters,
        uri=f"/sap/bc/adt/oo/classes/{class_name}/includes",
        method="POST",
        body=body,
        params={"lockHandle": lock_handle},
    )

    if 200 <= response.status_code <= 300:
        return True
    else:
        raise Exception(
            f"{response.status_code} - Failed to create testclass for {class_name}\n{response.text}"
        )



import xml.etree.ElementTree as et

from ..compat_typing import List, TypedDict
from ..api.xml_namespaces import XML_NAMESPACES
from ..http_request import HttpRequestParameters, request


class UnittestFlags:
    def __init__(
        self,
        harmless: bool = True,
        dangerous: bool = False,
        critical: bool = False,
        short: bool = True,
        medium: bool = False,
        long: bool = False,
    ):
        self.harmless = harmless
        self.dangerous = dangerous
        self.critical = critical
        self.short = short
        self.medium = medium
        self.long = long


class UnitTestAlert(TypedDict):
    title: str
    kind: str
    severity: str
    details: List[str]


def _parse_alerts(xml_text: str) -> List[UnitTestAlert]:
    root = et.fromstring(xml_text)
    errors = []
    alerts = root.findall(".//alert", XML_NAMESPACES)
    for alert in alerts:
        a = {}
        title = alert.find(".//title", XML_NAMESPACES)
        a["title"] = title.text if title is not None else ""
        a["kind"] = alert.attrib.get("kind", "")
        a["severity"] = alert.attrib.get("severity", "")
        stack = alert.find(".//stack", XML_NAMESPACES)
        a["stack"] = stack if stack is not None else []
        details = []
        for detail in alert.findall("./details/detail", XML_NAMESPACES):
            detail_str = "\n".join(
                x.attrib["text"]
                for x in detail.iter()
                if "text" in x.attrib is not None
            )
            details.append(detail_str)
        a["details"] = details
        errors.append(a)
    return errors


def run_unit_test(
    http_request_parameters: HttpRequestParameters,
    object_uri: str,
    unit_test_flags: UnittestFlags = UnittestFlags(),
):
    body = f"""
        <?xml version="1.0" encoding="UTF-8"?>
        <aunit:runConfiguration xmlns:aunit="http://www.sap.com/adt/aunit">
        <external>
            <coverage active="false"/>
        </external>
        <options>
            <uriType value="semantic"/>
            <testDeterminationStrategy sameProgram="true" assignedTests="false"/>
            <testRiskLevels harmless="{str(unit_test_flags.harmless).lower()}" dangerous="{str(unit_test_flags.dangerous).lower()}" critical="{str(unit_test_flags.critical).lower()}"/>
            <testDurations short="{str(unit_test_flags.short).lower()}" medium="{str(unit_test_flags.medium).lower()}" long="{str(unit_test_flags.long).lower()}"/>
            <withNavigationUri enabled="true"/>    
        </options>
        <adtcore:objectSets xmlns:adtcore="http://www.sap.com/adt/core">
            <objectSet kind="inclusive">
            <adtcore:objectReferences>
                <adtcore:objectReference adtcore:uri="{object_uri}"/>
            </adtcore:objectReferences>
            </objectSet>
        </adtcore:objectSets>
        </aunit:runConfiguration>
        """

    response = request(
        http_request_parameters=http_request_parameters,
        uri="/sap/bc/adt/abapunit/testruns",
        method="POST",
        body=body,
        params={},
        content_type="application/xml",
    )
    if response.status_code == 200:
        alerts = _parse_alerts(response.text)
        return alerts

    else:
        raise Exception(
            f"{response.status_code} - Failed to run unit test for {object_uri}\n{response.text}"
        )



XML_NAMESPACES = {
    "chkl": "http://www.sap.com/abapxml/checklist",
    "atom": "http://www.w3.org/2005/Atom",
    "adtcore": "http://www.sap.com/adt/core",
    "exc": "http://www.sap.com/abapxml/types/communicationframework",
    "asx": "http://www.sap.com/abapxml",
    "aunit": "http://www.sap.com/adt/aunit",
    "chkrun": "http://www.sap.com/adt/checkrun",
    "abapsource": "http://www.sap.com/adt/abapsource",
}


import pytest
from abap_adt_py.adt_client import AdtClient

import pytest

TEST_DELETE_ORDER = 10


@pytest.mark.order(TEST_DELETE_ORDER)
def test_delete_report(client: AdtClient, test_report: str):
    uri = f"/sap/bc/adt/programs/programs/{test_report}"
    lock_handle = client.lock(uri)
    client.delete(uri, lock_handle)


@pytest.mark.order(TEST_DELETE_ORDER)
def test_delete_class(client: AdtClient, test_class: str):
    uri = f"/sap/bc/adt/oo/classes/{test_class}"
    lock_handle = client.lock(uri)
    client.delete(uri, lock_handle)
    client.unlock(uri, lock_handle)


# @pytest.mark.order(TEST_DELETE_ORDER)
# def test_delete_table(client: AdtClient, test_table: str):
#     uri = f"/sap/bc/adt/ddic/tables/{test_table}"
#     lock_handle = client.lock(uri)
#     client.delete(uri, lock_handle)


# @pytest.mark.order(TEST_DELETE_ORDER)
# def test_delete_function_module(client: AdtClient, test_func_mod: str):
#     uri = f"/sap/bc/adt/functions/groups/{test_func_mod}"  # Usually FMs are subnodes of FUGRs
#     lock_handle = client.lock(uri)
#     client.delete(uri, lock_handle)


# @pytest.mark.order(TEST_DELETE_ORDER)
# def test_delete_function_group(client: AdtClient, test_fugr: str):
#     uri = f"/sap/bc/adt/functions/groups/{test_fugr}"
#     lock_handle = client.lock(uri)
#     client.delete(uri, lock_handle)


# @pytest.mark.order(TEST_DELETE_ORDER)
# def test_delete_message_class(client: AdtClient, test_msag: str):
#     uri = f"/sap/bc/adt/messages/{test_msag}"
#     lock_handle = client.lock(uri)
#     client.delete(uri, lock_handle)


# @pytest.mark.order(TEST_DELETE_ORDER)
# def test_delete_dcl_source(client: AdtClient, test_dcls: str):
#     uri = f"/sap/bc/adt/authorization/dcls/{test_dcls}"
#     lock_handle = client.lock(uri)
#     client.delete(uri, lock_handle)


# @pytest.mark.order(TEST_DELETE_ORDER)
# def test_delete_ddls_view(client: AdtClient, test_ddls: str):
#     uri = f"/sap/bc/adt/cds/ddls/{test_ddls}"
#     lock_handle = client.lock(uri)
#     client.delete(uri, lock_handle)


# @pytest.mark.order(TEST_DELETE_ORDER)
# def test_delete_ddlx_extension(client: AdtClient, test_ddlx: str):
#     uri = f"/sap/bc/adt/cds/ddlextensions/{test_ddlx}"
#     lock_handle = client.lock(uri)
#     client.delete(uri, lock_handle)


# @pytest.mark.order(TEST_DELETE_ORDER)
# def test_delete_data_element(client: AdtClient, test_dtel: str):
#     uri = f"/sap/bc/adt/ddic/dataelements/{test_dtel}"
#     lock_handle = client.lock(uri)
#     client.delete(uri, lock_handle)



import requests
from .compat_typing import Literal, TypedDict


class HttpRequestParameters(TypedDict):
    host: str
    csrf_token: str
    statefulness: Literal["stateless", "stateful"]
    request_number: int
    session: requests.Session


def request(
    http_request_parameters: HttpRequestParameters,
    uri: str,
    method: Literal["GET", "POST", "PUT", "DELETE"],
    body: str,
    params: dict,
    content_type: str = "application/xml",
) -> requests.Response:

    config = {
        "params": params,
        "headers": {
            "Accept": "*/*",
            "Cache-Control": "no-cache",
            "x-csrf-token": http_request_parameters["csrf_token"],
            "X-sap-adt-sessiontype": http_request_parameters["statefulness"],
            "content-type": content_type,
        },
        "url": http_request_parameters["host"] + uri,
        "data": body,
    }

    session = http_request_parameters["session"]
    if method == "POST":
        response = session.post(**config)
    elif method == "GET":
        response = session.get(**config)
    elif method == "PUT":
        response = session.put(**config)
    elif method == "DELETE":
        response = session.delete(**config)
    else:
        raise ValueError(f"Unsupported method: {method}")
    return response



import xml.etree.ElementTree as et
from ..compat_typing import TypedDict, List
from ..http_request import HttpRequestParameters, request
from .xml_namespaces import XML_NAMESPACES


class LinkInfo(TypedDict):
    rel: str
    href: str


class ComponentInfo(TypedDict, total=False):
    name: str
    type: str
    links: List[LinkInfo]
    level: str
    clif_name: str
    visibility: str


class ClassStructureResult(TypedDict):
    name: str
    xml_base: str
    visibility: str
    final: bool
    type: str
    links: List[LinkInfo]
    components: List[ComponentInfo]


def _parse_class_structure_response(xml_content: str) -> ClassStructureResult:
    root = et.fromstring(xml_content)

    all_attrs = root.attrib

    class_info: ClassStructureResult = {
        "name": "",
        "xml_base": "",
        "visibility": "",
        "final": False,
        "type": "",
        "links": [],
        "components": [],
    }

    attr_mapping = {
        "{http://www.sap.com/adt/core}name": "name",
        "{http://www.w3.org/XML/1998/namespace}base": "xml_base",
        "visibility": "visibility",
        "{http://www.sap.com/adt/core}type": "type",
    }
    for xml_attr, struct_key in attr_mapping.items():
        if xml_attr in all_attrs:
            if struct_key == "final":
                class_info[struct_key] = all_attrs[xml_attr] == "true"
            else:
                class_info[struct_key] = all_attrs[xml_attr]

    if "final" in all_attrs:
        class_info["final"] = all_attrs["final"] == "true"

    for link in root.findall("atom:link", XML_NAMESPACES):
        class_info["links"].append(
            {"rel": link.get("rel", ""), "href": link.get("href", "")}
        )

    for element in root.findall(".//abapsource:objectStructureElement", XML_NAMESPACES):
        all_element_attrs = element.attrib

        element_info: ComponentInfo = {
            "name": "",
            "type": "",
            "links": [],
        }
        for attr_name, attr_value in all_element_attrs.items():
            if attr_value in ["true", "false"]:
                element_info[attr_name] = attr_value == "true"
            else:
                if attr_name in attr_mapping:
                    attr_name = attr_mapping[attr_name]
                element_info[attr_name] = attr_value

        for link in element.findall("atom:link", XML_NAMESPACES):
            element_info["links"].append(
                {"rel": link.get("rel", ""), "href": link.get("href", "")}
            )
        class_info["components"].append(element_info)

    return class_info


def object_structure(http_request_parameters: HttpRequestParameters, object_uri: str):
    response = request(
        http_request_parameters,
        uri=f"{object_uri}/objectstructure",
        params={"version": "active", "withShortDescriptions": True},
        body="",
        method="GET",
        content_type="application/*",
    )
    if 200 <= response.status_code < 300:
        content = _parse_class_structure_response(response.text)
        return content
    else:
        raise Exception(
            f"{response.status_code} - Failed get object structure for {object_uri}\n{response.text}"
        )



from ..http_request import HttpRequestParameters, request


def login(http_request_parameters: HttpRequestParameters) -> str:

    response = request(
        http_request_parameters=http_request_parameters,
        uri="/sap/bc/adt/compatibility/graph",
        method="GET",
        body="",
        params={},
    )

    if response.status_code == 200:
        csrf_token = response.headers.get("x-csrf-token")
        if csrf_token is None:
            raise Exception("CSRF token not found in response headers.")
        return csrf_token
    else:
        raise Exception(f"{response.status_code} - Login failed.\n{response.text}")



from xml.etree import ElementTree as et
from .compat_typing import Optional, Dict, List
from .api.xml_namespaces import XML_NAMESPACES


def _strip_namespace(name: str) -> str:
    if name.startswith("{"):
        return name.split("}", 1)[1]
    if ":" in name:
        return name.split(":", 1)[1]
    return name


def _et_to_attributes_dict(element: Optional[et.Element]) -> Dict[str, str]:
    cleaned = (
        {_strip_namespace(key): value for key, value in element.attrib.items()}
        if element is not None
        else {}
    )
    return cleaned


def find_xml_elements_attributes(xml_text: str, tag_name: str) -> List[Dict[str, str]]:
    root = et.fromstring(xml_text)
    elements = root.findall(tag_name, XML_NAMESPACES)
    processed_elements = []
    for element in elements:
        cleaned = _et_to_attributes_dict(element)
        processed_elements.append(cleaned)
    return processed_elements


def find_xml_element_attributes(xml_text: str, tag_name: str) -> Dict[str, str]:
    root = et.fromstring(xml_text)
    element = root.find(tag_name, XML_NAMESPACES)
    element = _et_to_attributes_dict(element)
    return element


def find_xml_element_text(xml_text: str, tag_name: str):
    root = et.fromstring(xml_text)
    el = root.find(tag_name, XML_NAMESPACES)
    return "".join(el.itertext()).strip() if el is not None else ""
